#include "femSolver.h"
#include "femProblem.h"
#include "femPartition.h"
#include "femPart.h"
#include "femDirichlet.h"
#include "femNeumann.h"
#include "femGroup.h"
#include "femEntity.h"
#include "femElementSet.h"
#include "femMedium.h"
#include "femMaps.h"
#include "OpenMP.h"
#include "femGmsh.h"
#include <iomanip>
#include <algorithm>
#include <Eigen/IterativeLinearSolvers>
#ifdef FEM_USE_MKL
#include <Eigen/PardisoSupport>
#endif

namespace fem
{

Solver::Solver(Problem &_pbl) : pbl(_pbl),
                                partition(nullptr),
                                part_Mfree(nullptr),
                                part_Mfixed(nullptr),
                                linear_solver("ldlt")
{
    dofs.add("X");
    dofs.add("Y");
    dofs.add("Z");
    bcs.add("FREE");
    bcs.add("PRESCRIBED");

    std::cout << "#threads = " << OpenMP::get_max_threads() << '\n';
}

Solver::~Solver()
{
    if(debug_level()>0)
        std::cout << "~Solver()\n";
    if (partition)
        delete partition;
    if (part_Mfree)
        delete part_Mfree;
    if (part_Mfixed)
        delete part_Mfixed;

    // free sparse K
    for (size_t i = 0; i < K.size(); ++i)
        for (size_t j = 0; j < K[i].size(); ++j)
            delete K[i][j];

    // free f, X
    for (size_t i = 0; i < f.size(); ++i)
        delete f[i];
    for (size_t i = 0; i < f.size(); ++i)
        delete X[i];
}

/// solve the problem

void Solver::solve()
{
    std::cout << pbl;

    create_partition();
    fill_partition();
    fill_lhs();
    fill_rhs();
    apply_bc();
    solve_static();

    std::cout << "Summary of CPU times:\n";

    std::ios cout_state(nullptr);
    cout_state.copyfmt(std::cout);

    std::cout << std::fixed << std::setprecision(1);
    for (auto const &[name, timer] : timers)
    {
        std::cout << '\t' << std::setw(15) << std::left << name << ":"
                  << std::setw(10) << std::right << timer.read() << '\n';
    }

    std::cout.copyfmt(cout_state);
}

// create a partition of the dofs

void Solver::create_partition()
{
    partition = new Partition();
    part_Mfree = new Part(*partition, dofs["X"] | dofs["Y"] | dofs["Z"], bcs["FREE"]);
    part_Mfixed = new Part(*partition, dofs["X"] | dofs["Y"] | dofs["Z"], bcs["PRESCRIBED"]);
    std::cout << *partition;
}

// sort nodes & dofs into the defined partition
// output:
//     dofFixed: is a dof fixed or free?
//     dofPart: in which part is a given dof of a given node?
//     dofIdx: what is its number in its part?
//     partSizes: what are the size of the parts in the partition?

void Solver::fill_partition()
{
    timers["fill_partition"].start();

    std::cout << "fill_partition:\n";

    // pbl.nodeTags (size: nb of nodes): gives the gmsh tag for each node
    // pbl.nodemap (size: max gmsh tag): gives the global node idx (position in nodeTags)
    //                                   from the gmsh tag
    //                                   (some space could be unused if non-contiguous numbering)
    //
    // dofPart (size: nb of nodes x size of dofs): for each node/dof, gives its partition number
    // dofIdx  (size: nb of nodes x size of dofs): for each node/dof, gives its index in its partition

    // fill the 'dofFixed' array:
    //             T  X  Y  Z
    //  node 0   [ x  x  x  x ]
    //  node 1   [ x  x  x  x ]  where 'x' is PRESCRIBED or FREE
    //  node 2   [ x  x  x  x ]
    //  ...

    dofFixed.setConstant(pbl.nodeTags.size(), dofs.size(), 0); // = false == FREE by default

    for (auto &bc : pbl.dirichlets)
    {
        DOF bc_dof = dofs[bc->m_dofname];
        for (auto &entity : bc->group->entities)
            for (auto &elems : entity->esets)
                for (size_t i = 0; i < elems->etags.size(); ++i)
                    for (int j = 0; j < elems->type_nnods; ++j)
                    {
                        // size_t n = elems->ntags_v[i * elems->type_nnods + j];
                        size_t n = elems->ntags(i, j);
                        size_t gidx = pbl.nodemap[n];
                        dofFixed(gidx, bc_dof.index()) = 1; // PRESCRIBED
                    }
    }

    Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
    // std::cout << "dofFixed =\n"
    //           << dofFixed.format(fmt) << '\n';

    // fill dofPart (default:-1 => no partition assigned)
    //             T  X  Y  Z
    //  node 0   [ x  x  x  x ]
    //  node 1   [ x  x  x  x ]  where 'x' is the partition of the dof
    //  node 2   [ x  x  x  x ]
    //  ...

    dofPart.setConstant(pbl.nodeTags.size(), dofs.size(), -1);

    for (size_t i = 0; i < pbl.nodeTags.size(); ++i)
        for (size_t j = 0; j < dofs.size(); ++j) // TODO: iterator
        {
            DOF dof = dofs[j];
            DOF bc = dofFixed(i, j) ? bcs["PRESCRIBED"] : bcs["FREE"];
            for (auto &p : partition->parts)
                if ((p->dofmask & dof) && (p->bcmask & bc))
                {
                    dofPart(i, j) = p->idx;
                    break;
                }
        }

    // std::cout << "dofPart =\n"
    //           << dofPart.format(fmt) << '\n';

    // count nodes in partitions
    int nnods_Mfree = (dofPart == part_Mfree->idx).count();
    std::cout << "\tfree M dofs = " << nnods_Mfree << '\n';
    int nnods_Mfixed = (dofPart == part_Mfixed->idx).count();
    std::cout << "\tfixed M dofs = " << nnods_Mfixed << std::endl;

    // renumber nodes in each partition
    // fill dofIdx (default:-1 => no part assigned)
    //             T  X  Y  Z
    //  node 0   [ x  x  x  x ]
    //  node 1   [ x  x  x  x ]  where 'x' is the index of the dof in its partition
    //  node 2   [ x  x  x  x ]
    //  ...
    // fill also 'partSizes' (nb of dofs in each part)

    dofIdx.setConstant(pbl.nodeTags.size(), dofs.size(), 0);
    partSizes.resize(partition->parts.size(), 0);

    for (size_t i = 0; i < pbl.nodeTags.size(); ++i)
        for (size_t j = 0; j < dofs.size(); ++j) // TODO: iterator
        {
            int p = dofPart(i, j);
            dofIdx(i, j) = partSizes[p];
            partSizes[p]++;
        }

    // std::cout << "dofIdx =\n"
    //           << dofIdx.format(fmt) << '\n';
    // std::cout << "partSizes = " << partSizes << std::endl;

    // TODO: check that all dofs are assigned?
    timers["fill_partition"].stop();
    std::cout << "\telasped = " << timers["fill_partition"].read() << "s\n";
}

/// fill left-hand side of the system of equations

void Solver::fill_lhs()
{
    timers["fill_lhs"].start();

    std::cout << "assembly of LHS...\n";

    // -- prepare structural matrices
    //
    //  Kijv[nparts][nparts], Cijv[nparts][nparts]:
    //       "ijv" versions of K, C and M used during assembly

    std::vector<std::vector<std::vector<Eigen::Triplet<double>> *>> Kijv(partition->parts.size());
    for (size_t i = 0; i < partition->parts.size(); ++i)
    {
        Kijv[i].resize(partition->parts.size());
        for (size_t j = 0; j < partition->parts.size(); ++j)
            Kijv[i][j] = new std::vector<Eigen::Triplet<double>>;
    }

    // -- pre-compute sub-matrices structures
    std::vector<std::vector<size_t>> ntriplets(partition->parts.size(), std::vector<size_t>(partition->parts.size(), 0));

    std::vector<DOF> dofs_M = {dofs["X"], dofs["Y"], dofs["Z"]};
    for (auto medium : pbl.media)
        for (auto entity : medium->group->entities)
            for (auto elems : entity->esets)
            {
                const int ndim = elems->type_dim;
                for (size_t ie = 0; ie < elems->etags.size(); ++ie)
                {
                    // brutal assembly (needs optimisation)
                    for (int i = 0; i < elems->type_nnods; ++i)
                    {
                        int i_tag = elems->ntags(ie, i);   // gmsh tag of node i
                        int i_gidx = pbl.nodemap[i_tag];   // global index of node i
                        auto i_part = dofPart.row(i_gidx); // parts of dofs of node i
                        for (int j = 0; j < elems->type_nnods; ++j)
                        {
                            int j_tag = elems->ntags(ie, j);   // gmsh tag of node j
                            int j_gidx = pbl.nodemap[j_tag];   // global index of node j
                            auto j_part = dofPart.row(j_gidx); // parts of dofs of node j

                            for (int ni = 0; ni < ndim; ++ni)
                            {
                                DOF &di = dofs_M[ni];
                                for (int nj = 0; nj < ndim; ++nj)
                                {
                                    DOF &dj = dofs_M[nj];

                                    int ip = i_part(di.index());
                                    int jp = j_part(dj.index());
                                    ntriplets[ip][jp]++;
                                }
                            }
                        }
                    }
                }
            }                    

    // reserve the correct amount of triplets 
    // (avoids reallocation in later push_backs)
    for (size_t i = 0; i < partition->parts.size(); ++i)
    {
        Kijv[i].resize(partition->parts.size());
        for (size_t j = 0; j < partition->parts.size(); ++j)
        {
            // std::cout << "expecting " << ntriplets[i][j] 
            //           << " triplets for submatrix " << i << "," << j << '\n';
            Kijv[i][j]->reserve(ntriplets[i][j]);
        }
    }

    // -- compute elemental matrices

    for (auto medium : pbl.media)
        for (auto entity : medium->group->entities)
            for (auto elems : entity->esets)
            {
                std::cout << "\tprocessing " << *elems << '\n';
                size_t nelems = elems->etags.size();
                std::cout << "\t\tnelems = " << nelems << '\n';
                // integration points
                std::vector<double> localCoord;
                std::vector<double> weights;
                gmsh::model::mesh::getIntegrationPoints(elems->type_id,
                                                        pbl.quadrature,
                                                        localCoord,
                                                        weights);
                size_t nGP = weights.size();
                std::cout << "\t\tnGP = " << nGP << '\n';
                // jacobians
                std::vector<double> jacobians_v;
                std::vector<double> determinants_v;
                std::vector<double> coord_v;
                gmsh::model::mesh::getJacobians(elems->type_id, localCoord,
                                                jacobians_v, determinants_v,
                                                coord_v, entity->tag);
                // reshape (map) into matrices
                JacoMap jacobians(jacobians_v, nelems, nGP);
                Eigen::Map<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>>
                    determinants(&determinants_v[0], nelems, nGP);
                CoordMap coord(coord_v, nelems, nGP);

                // print GP data
                // for (size_t i = 0; i < nelems; ++i)
                // {
                //     std::cout << "\telem #" << i << " (tag=" << elems->ntags[i] << "):\n";
                //     for (size_t j = 0; j < nGP; ++j)
                //     {
                //         std::cout << "\t\tGP #" << j + 1 << '/' << nGP << ":\n";
                //         Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t\t\t\t[", "]");
                //         std::cout << "\t\t\tJ =\n"
                //                   << jacobians(i, j).format(fmt) << '\n';
                //         std::cout << "\t\t\tdet = " << jacobians(i, j).determinant()
                //                   << " (gmsh = " << determinants(i, j) << ")\n";
                //         std::cout << "\t\t\tcoord =\n"
                //                   << coord(i, j).format(fmt) << '\n';
                //     }
                // }

                // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                // std::cout << "determinants =\n"
                //         << determinants.format(fmt) << '\n';
                // std::cout << "determinants_v =\n" << determinants_v << '\n';
                // std::cout << "coord_v =\n" << coord_v << '\n';
                // std::cout << "coord_v.size() =" << coord_v.size() << '\n';

                // basis functions
                int numComponents; // unused
                std::vector<double> basisF_v;
                int numOrientations; // unused
                gmsh::model::mesh::getBasisFunctions(elems->type_id,
                                                     localCoord, "Lagrange",
                                                     numComponents, // unused
                                                     basisF_v,
                                                     numOrientations); // unused
                Eigen::Map<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>>
                    basisF(&basisF_v[0], nGP, elems->type_nnods);
                // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                // std::cout << "basisF =\n"
                //         << basisF.format(fmt) << '\n';
                // std::cout << "basisF_v =\n"
                //         << basisF_v << '\n';
                // exit(1);

                // basis functions (derivatives)
                std::vector<double> basisDF_v;
                gmsh::model::mesh::getBasisFunctions(elems->type_id,
                                                     localCoord, "GradLagrange",
                                                     numComponents, // reused
                                                     basisDF_v,
                                                     numOrientations); // reused
                DShapeMap basisDF(basisDF_v, nGP, elems->type_nnods);
                // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                // std::cout << "basisDF =\n"
                //           << basisDF(0).format(fmt) << '\n';
                // std::cout << "basisDF_v =\n"
                //           << basisDF_v << '\n';
                // exit(1);

                #ifdef _MSC_VER
                #define OPENMP_INT int
                #else
                #define OPENMP_INT size_t
                #endif

                // loop over the elements
                #pragma omp parallel for
                for (OPENMP_INT ie = 0; ie < elems->etags.size(); ++ie)
                {
                    // std::cout << "\tprocessing element #" << elems->etags[ie] << '\n';
                    const int ndim = elems->type_dim;

                    // compute elemental matrices
                    Eigen::MatrixXd Ke_MM = Eigen::MatrixXd::Zero(ndim * elems->type_nnods, ndim * elems->type_nnods);

                    // loop over Gauss points
                    for (size_t igp = 0; igp < nGP; ++igp)
                    {
                        auto N_T = basisF.row(igp); // matrix of shape functions; shape=(1,elems.type_nnods)
                        // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                        // std::cout << "N_T (igp=" << igp << ") = " << N_T.format(fmt) << '\n';
                        auto DN = basisDF(igp).block(0, 0, elems->type_nnods, ndim); // matrix of shp fct derivatives in 'xi' space; shape=(elems.type_nnods,3)
                        // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                        // std::cout << "DN (igp=" << igp << ") =\n"
                        //           << DN.format(fmt) << '\n';
                        auto J = jacobians(ie, igp).block(0, 0, ndim, ndim);
                        auto Jt = J.transpose();   // transposed jacobian matrix; shape=(3,3)
                        auto Jtinv = Jt.inverse(); // transposed jacobian inverse; shape=(3,3)
                        // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                        // std::cout << "Jt*Jtinv = " << (Jt*Jtinv).format(fmt) << '\n';
                        auto B_T = Jtinv * DN.transpose(); // matrix of shp fct derivatives in 'x' space; shape=(3, nnods_by_element)
                        // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                        // std::cout << "B_T = " << B_T.format(fmt) << '\n';
                        // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                        // std::cout << "basisF = " << basisF.format(fmt) << '\n';

                        Eigen::MatrixXd N_M(ndim, elems->type_nnods * ndim);
                        for (int j = 0; j < elems->type_nnods; ++j)
                        {
                            // double v = basisF(igp, j);
                            double v = N_T(j);
                            // std::cout << "v = " << v << '\n';
                            N_M.block(0, ndim * j, ndim, ndim) = Eigen::MatrixXd::Identity(ndim, ndim) * v;
                        }
                        // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                        // std::cout << "N_M = " << N_M.format(fmt) << '\n';

                        // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                        // std::cout << "B_T = " << B_T.format(fmt) << '\n';

                        Eigen::MatrixXd B_M;

                        switch (ndim)
                        {
                        case 2:

                            B_M.resize(3, elems->type_nnods * ndim); // 3 en 2D; 6 en 3D
                            for (int j = 0; j < elems->type_nnods; ++j)
                            {
                                double ddx = B_T(0, j);
                                double ddy = B_T(1, j);
                                Eigen::MatrixXd Bs(3, 2);
                                Bs << ddx, 0.0,
                                    0.0, ddy,
                                    ddy, ddx;
                                B_M.block(0, ndim * j, 3, 2) = Bs;
                            }
                            break;
                        case 3:

                            B_M.resize(6, elems->type_nnods * ndim); // 3 en 2D; 6 en 3D
                            for (int j = 0; j < elems->type_nnods; ++j)
                            {
                                // double v = basisF(igp, j);
                                double ddx = B_T(0, j);
                                double ddy = B_T(1, j);
                                double ddz = B_T(2, j);
                                Eigen::MatrixXd Bs(6, 3);
                                Bs << ddx, 0.0, 0.0,
                                    0.0, ddy, 0.0,
                                    0.0, 0.0, ddz,
                                    ddy, ddx, 0.0,
                                    ddz, 0.0, ddx,
                                    0.0, ddz, ddy;
                                // std::cout << "v = " << v << '\n';
                                B_M.block(0, ndim * j, 6, 3) = Bs;
                            }
                            break;
                        default:
                            throw std::runtime_error("dimension " + std::to_string(ndim) + " not implemented");
                        }
                        // std::cout << "B_M = " << B_M.format(fmt) << '\n';

                        Ke_MM += B_M.transpose() * medium->hooke(ndim) * B_M * determinants(ie, igp) * weights[igp];
                        //
                        // exit(1);
                    }
                    // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                    // std::cout << "Ke_MM = \n"
                    //           << Ke_MM.format(fmt) << '\n';
                    // exit(1);

                    std::vector<DOF> dofs_M = {dofs["X"], dofs["Y"], dofs["Z"]};

                    // brutal assembly (needs optimisation)
                    #pragma omp critical
                    for (int i = 0; i < elems->type_nnods; ++i)
                    {
                        int i_tag = elems->ntags(ie, i);   // gmsh tag of node i
                        int i_gidx = pbl.nodemap[i_tag];   // global index of node i
                        auto i_part = dofPart.row(i_gidx); // parts of dofs of node i
                        auto i_idx = dofIdx.row(i_gidx);   // index of these dofs in their parts
                        for (int j = 0; j < elems->type_nnods; ++j)
                        {
                            int j_tag = elems->ntags(ie, j);   // gmsh tag of node j
                            int j_gidx = pbl.nodemap[j_tag];   // global index of node j
                            auto j_part = dofPart.row(j_gidx); // parts of dofs of node j
                            auto j_idx = dofIdx.row(j_gidx);   // index of these dofs in their parts

                            for (int ni = 0; ni < ndim; ++ni)
                            {
                                DOF &di = dofs_M[ni];
                                for (int nj = 0; nj < ndim; ++nj)
                                {
                                    DOF &dj = dofs_M[nj];
                                    // std::cout << "(" << di.index() << ',' << dj.index() << ")\n";
                                    // get vector of triplets
                                    auto K_sp = Kijv[i_part(di.index())][j_part(dj.index())];
                                    K_sp->push_back(Eigen::Triplet<double>(
                                        i_idx(di.index()),
                                        j_idx(dj.index()),
                                        Ke_MM(i * ndim + ni, j * ndim + nj)));
                                }
                            }
                        }
                    } // assembly

                } // loop over the elements
            }     // for (auto &elems : entity->esets)

    // create sparse K

    K.resize(partition->parts.size());
    for (size_t i = 0; i < partition->parts.size(); ++i)
    {
        K[i].resize(partition->parts.size());
        for (size_t j = 0; j < partition->parts.size(); ++j)
        {
            K[i][j] = new Eigen::SparseMatrix<double>(partSizes[i], partSizes[j]);
            K[i][j]->setFromTriplets(Kijv[i][j]->begin(), Kijv[i][j]->end()); // TODO: use std::list!
        }
    }

    // free triplets
    for (size_t i = 0; i < Kijv.size(); ++i)
        for (size_t j = 0; j < Kijv[i].size(); ++j)
            delete Kijv[i][j];

    timers["fill_lhs"].stop();
    std::cout << "\telasped = " << timers["fill_lhs"].read() << "s\n";
}

/// compute right-hand side 'f' (sources)

void Solver::fill_rhs()
{
    timers["fill_rhs"].start();

    std::cout << "assembly of RHS...\n";

    // create RHS
    f.resize(partition->parts.size());
    for (size_t i = 0; i < partition->parts.size(); ++i)
    {
        f[i] = new Eigen::VectorXd(partSizes[i]);
        f[i]->setZero();
    }

    // sources
    // ...

    // std::cout << "Neumann BCs...\n";

    for (auto src : pbl.neumanns)
    {
        DOF bc_dof = dofs[src->m_dofname];
        if (src->group)
        {
            for (auto entity : src->group->entities) // each source has a target group of entities
                for (auto elems : entity->esets)     // each entity has a several sets of elements
                {
                    size_t nelems = elems->etags.size();
                    // integration points
                    std::vector<double> localCoord;
                    std::vector<double> weights;
                    gmsh::model::mesh::getIntegrationPoints(elems->type_id,
                                                            pbl.quadrature,
                                                            localCoord,
                                                            weights);
                    size_t nGP = weights.size();
                    // jacobians
                    std::vector<double> jacobians_v;
                    std::vector<double> determinants_v;
                    std::vector<double> coord_v;
                    gmsh::model::mesh::getJacobians(elems->type_id, localCoord,
                                                    jacobians_v, determinants_v,
                                                    coord_v, entity->tag);
                    // reshape (map) into matrices
                    JacoMap jacobians(jacobians_v, nelems, nGP);
                    Eigen::Map<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>>
                        determinants(&determinants_v[0], nelems, nGP);
                    CoordMap coord(coord_v, nelems, nGP);
                    // basis functions
                    int numComponents; // unused
                    std::vector<double> basisF_v;
                    int numOrientations; // unused
                    gmsh::model::mesh::getBasisFunctions(elems->type_id,
                                                         localCoord, "Lagrange",
                                                         numComponents, // unused
                                                         basisF_v,
                                                         numOrientations); // unused
                    Eigen::Map<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>>
                        basisF(&basisF_v[0], nGP, elems->type_nnods);

                    // loop over the elements
                    for (size_t ie = 0; ie < elems->etags.size(); ++ie)
                    {
                        Eigen::VectorXd fe = Eigen::VectorXd::Zero(elems->type_nnods);
                        // loop over Gauss points
                        for (size_t igp = 0; igp < nGP; ++igp)
                        {
                            auto N_T = basisF.row(igp); // matrix of shape functions; shape=(1,elems.type_nnods)
                            fe += src->m_val * N_T * determinants(ie, igp) * weights[igp];
                        }

                        // brutal assembly (needs optimisation)
                        for (int i = 0; i < elems->type_nnods; ++i)
                        {
                            int i_tag = elems->ntags(ie, i);                   // gmsh tag of node i
                            int i_gidx = pbl.nodemap[i_tag];                   // global index of node i
                            auto i_part = dofPart.row(i_gidx)[bc_dof.index()]; // parts of dofs of node i
                            auto i_idx = dofIdx.row(i_gidx)[bc_dof.index()];   // index of these dofs in their parts
                            (*f[i_part])[i_idx] += fe[i];
                        }
                    }
                }
        } // if (src->group)
        else
        {
            // neumann applied on a single node
            int i_tag = src->m_nodetag;                        // gmsh tag of node i
            int i_gidx = pbl.nodemap[i_tag];                   // global index of node i
            auto i_part = dofPart.row(i_gidx)[bc_dof.index()]; // parts of dofs of node i
            auto i_idx = dofIdx.row(i_gidx)[bc_dof.index()];   // index of these dofs in their parts
            (*f[i_part])[i_idx] += src->m_val;
        }
    }

    timers["fill_rhs"].stop();
    std::cout << "\telasped = " << timers["fill_rhs"].read() << "s\n";

}

// create the X vector (containing dofs)
// and prescribe dirichlet boundary conditions
void Solver::apply_bc()
{
    timers["apply_bc"].start();
    std::cout << "applying dirichlet boundary conditions...\n";

    // create vectors of dofs
    X.resize(partition->parts.size());
    for (size_t i = 0; i < partition->parts.size(); ++i)
    {
        X[i] = new Eigen::VectorXd(partSizes[i]);
        X[i]->setZero();
    }

    // apply boundary conditions

    for (auto &bc : pbl.dirichlets)
    {
        DOF bc_dof = dofs[bc->m_dofname];
        for (auto &entity : bc->group->entities)
            for (auto &elems : entity->esets)
                for (size_t i = 0; i < elems->etags.size(); ++i)
                    for (int j = 0; j < elems->type_nnods; ++j)
                    {
                        // size_t n = elems->ntags_v[i * elems->type_nnods + j];
                        size_t n = elems->ntags(i, j);
                        size_t gidx = pbl.nodemap[n];
                        int part = dofPart(gidx, bc_dof.index());
                        int idx = dofIdx(gidx, bc_dof.index());
                        (*X[part])[idx] = bc->m_val;
                    }
    }

    timers["apply_bc"].stop();
    std::cout << "\telasped = " << timers["apply_bc"].read() << "s\n";
}

void Solver::solve_static()
{
    timers["solve_static"].start();

    std::cout << "solving linear system...\n";

    // create vectors of tmp_rhs
    std::vector<Eigen::VectorXd *> tmp_rhs(partition->parts.size());
    for (size_t i = 0; i < partition->parts.size(); ++i)
    {
        tmp_rhs[i] = new Eigen::VectorXd(partSizes[i]);
        tmp_rhs[i]->setZero();
    }

    // system: (F="free" ; P="prescribed")
    //     [ K_FF K_FP ] [ x_F ] = [ f_F ]
    //     [ K_PF K_PP ] [ x_P ]   [ f_P ]
    // construct right-hand sides
    //      [ K_FF ][ x_F ] + [ K_FP ][ x_P ] = [ f_F ]
    //   => [ K_FF ][ x_F ] = [ f_F ] - [ K_FP ][ x_P ]
    //                        <-------tmp_rhs_F------->
    //   (same procedure for both thermal and mechanical problems)

    size_t MF = part_Mfree->idx;
    size_t MP = part_Mfixed->idx;

    // mechanical
    *tmp_rhs[MF] = *f[MF] - (*K[MF][MP]) * (*X[MP]);

    // mechanical problem: [ x_F ] = [ K_FF ]^(-1) [ tmp_rhs_F ]
    std::cout << "\t#unknowns = " << K[MF][MF]->rows() << '\n';
    std::cout << "\t#threads = " << OpenMP::get_max_threads() << '\n';

    if (linear_solver == "cg")
    {
        // --conjugate gradient
        std::cout << "\tusing Eigen::ConjugateGradient (linear_solver=\""+linear_solver+"\")\n";
        Eigen::ConjugateGradient<Eigen::SparseMatrix<double>, Eigen::Lower | Eigen::Upper> solver;
        solver.compute(*K[MF][MF]);
        if(solver.info()!=Eigen::Success) 
            throw std::runtime_error("solver.compute() returned "+std::to_string(solver.info()));
        *X[MF] = solver.solve(*tmp_rhs[MF]);
        std::cout << "\t#iterations:     " << solver.iterations() << std::endl;
        std::cout << "\testimated error: " << solver.error() << std::endl;
        if(solver.info()!=Eigen::Success) 
            throw std::runtime_error("solver.solve() returned "+std::to_string(solver.info()));
    }
    else if (linear_solver == "ldlt")
    {
        // --direct solver (LDLT)
        std::cout << "\tusing Eigen::SimplicialLDLT (linear_solver=\""+linear_solver+"\")\n";
        Eigen::SimplicialLDLT<Eigen::SparseMatrix<double>> solver;
        solver.compute(*K[MF][MF]);
        if(solver.info()!=Eigen::Success) 
            throw std::runtime_error("solver.compute() returned "+std::to_string(solver.info()));
        *X[MF] = solver.solve(*tmp_rhs[MF]); // use the factorization to solve for the given right-hand side
        if(solver.info()!=Eigen::Success) 
            throw std::runtime_error("solver.solve() returned "+std::to_string(solver.info()));
    }
#ifdef FEM_USE_MKL
    else if (linear_solver == "pardiso")
    {
        // -- symmetric pardiso (LDLT)
        std::cout << "\tusing Eigen::PardisoLDLT (linear_solver=\""+linear_solver+"\")\n";
        Eigen::PardisoLDLT<Eigen::SparseMatrix<double>> solver;
        solver.compute(*K[MF][MF]);
        if(solver.info()!=Eigen::Success) 
            throw std::runtime_error("solver.compute() returned "+std::to_string(solver.info()));
        *X[MF] = solver.solve(*tmp_rhs[MF]); // use the factorization to solve for the given right-hand side
        if(solver.info()!=Eigen::Success) 
            throw std::runtime_error("solver.solve() returned "+std::to_string(solver.info()));
    }
#endif // FEM_USE_MKL
    else
    {
        throw std::runtime_error("unknown linear_solver: "+linear_solver);
    }

    // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
    // std::cout << "X[MF] = \n"
    //           << (*X[MF]).format(fmt) << '\n';

    // compute reactions: [ f_P ] = [ K_PF ][ x_F ] + [ K_PP ][ x_P ]
    //    values written previously in [ f_P ] during source / Neuman-BCs
    //    treatment are overwritten

    *f[MP] = (*K[MP][MF]) * (*X[MF]) + (*K[MP][MP]) * (*X[MP]);

    // std::cout << "f[MP] = \n"
    //           << (*f[MP]).format(fmt) << '\n';

    // energy: work of ext forces (using their "final" value - thus twice the int energy)
    double workF = (*f[MP]).dot(*X[MP]) +  (*f[MF]).dot(*X[MF]);
    //std::cout << "workF = " << workF << '\n';
    int_energy = workF/2;

    // free vectors of tmp_rhs
    for (size_t i = 0; i < tmp_rhs.size(); ++i)
        delete tmp_rhs[i];

    timers["solve_static"].stop();
    std::cout << "\telasped = " << timers["solve_static"].read() << "s\n";
}

// strain and stresses (epsilon, sigma) per node, per element
// (discontinuous across element boundaries)

void Solver::compute_sig_epl(std::vector<size_t> &etags_all,                            // tags of elements
                             std::vector<std::vector<double>> &sigma_all,               // values of stresses at the nodes of each element
                             std::vector<std::vector<double>> &epsilon_all,             // values of strains at the nodes of each element
                             std::map<size_t, std::vector<Eigen::VectorXd>> &sig_by_node, // value of all the stress vectors (Voigt) at a given node (key of the dict)
                             std::map<size_t, std::vector<Eigen::VectorXd>> &eps_by_node) // same for strain vector
{
    timers["compute_sig_epl"].start();

    std::cout << "computing strain and stresses...\n";

    // pre-allocate things
    std::map<size_t, size_t> nneighbours;
    size_t nelems=0;
    for (auto &medium : pbl.media)
        for (auto &entity : medium->group->entities)
        {
            for (auto &elems : entity->esets)
            {
                nelems+=elems->etags.size(); // accumulate nb of elems

                for (size_t ie = 0; ie < elems->etags.size(); ++ie)
                    for (int inod = 0; inod < elems->ntags.cols(); ++inod)
                    {
                        size_t ntag = elems->ntags(ie, inod);
                        nneighbours[ntag]++;
                    }
            }
        }
    // std::cout << "\texpecting " << nelems << " elements\n";
    etags_all.reserve(nelems);
    epsilon_all.reserve(nelems);
    sigma_all.reserve(nelems);

    size_t maxneigh=0;
    for(auto const &[ntag, nneigh] : nneighbours)
    {
        maxneigh = std::max(nneigh, maxneigh);
        eps_by_node[ntag].reserve(nneigh);
        sig_by_node[ntag].reserve(nneigh);
    }
    std::cout << "\tmax #neighbours = " << maxneigh << '\n';

    // loop over the media

    for (auto &medium : pbl.media)
        for (auto &entity : medium->group->entities)
            for (auto &elems : entity->esets)
            {
                // std::cout << "processing " << *elems << '\n';
                size_t nelems = elems->etags.size();

                // jacobians at nodes
                std::vector<double> jacobians_v;
                std::vector<double> determinants_v;
                std::vector<double> coord_v;
                gmsh::model::mesh::getJacobians(elems->type_id, elems->type_localNodeCoord,
                                                jacobians_v, determinants_v,
                                                coord_v, entity->tag);
                // reshape (map) into matrices
                JacoMap jacobians(jacobians_v, nelems, elems->type_nnods);
                Eigen::Map<Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>>
                    determinants(&determinants_v[0], nelems, elems->type_nnods);
                CoordMap coord(coord_v, nelems, elems->type_nnods);

                // basis functions (derivatives)
                int numComponents;   // unused
                int numOrientations; // unused
                std::vector<double> basisDF_v;
                gmsh::model::mesh::getBasisFunctions(elems->type_id,
                                                     elems->type_localNodeCoord,
                                                     "GradLagrange",
                                                     numComponents, // unused
                                                     basisDF_v,
                                                     numOrientations); // unused
                DShapeMap basisDF(basisDF_v, elems->type_nnods, elems->type_nnods);
                // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                // std::cout << "basisDF =\n"
                //           << basisDF(0).format(fmt) << '\n';
                // std::cout << "basisDF_v =\n"
                //           << basisDF_v << '\n';
                // exit(1);

                // loop over the elements
                //#pragma omp parallel
                for (size_t ie = 0; ie < elems->etags.size(); ++ie)
                {
                    // std::cout << "element " << ie+1 << "/" << elems->etags.size() << '\n';
                    size_t etag = elems->etags[ie];
                    const int ndim = elems->type_dim;

                    // compute the nodal displacement vectors ("displ")
                    //                   and the temperatures ("temp")
                    //                   of the current element
                    Eigen::VectorXd displ(elems->ntags.cols() * ndim);
                    // loop over the nodes
                    for (int inod = 0; inod < elems->ntags.cols(); ++inod)
                    {
                        size_t ntag = elems->ntags(ie, inod);
                        size_t gidx = pbl.nodemap[ntag];
                        for (int di = 0; di < ndim; ++di)
                        {
                            DOF dof = dofs[di];
                            int part = dofPart(gidx, dof.index());
                            int idx = dofIdx(gidx, dof.index());
                            displ[ndim * inod + di] = (*X[part])[idx];
                        }
                    }

                    std::vector<double> eps_nodes(9 * elems->ntags.cols(), 0.0); // strains at the nodes of the element in a 1d array
                    std::vector<double> sig_nodes(9 * elems->ntags.cols(), 0.0); // stresses at the nodes of the element in a 1d array

                    // loop over the nodes
                    // and compute the stresses/strains/heat_fluxes at the nodes

                    for (int inod = 0; inod < elems->ntags.cols(); ++inod)
                    {
                        size_t ntag = elems->ntags(ie, inod);
                        auto DN = basisDF(inod).block(0, 0, elems->type_nnods, ndim); // matrix of shp fct derivatives in 'xi' space; shape=(elems.type_nnods,3)
                        auto J = jacobians(ie, inod).block(0, 0, ndim, ndim);
                        auto Jt = J.transpose();           // transposed jacobian matrix; shape=(3,3)
                        auto Jtinv = Jt.inverse();         // transposed jacobian inverse; shape=(3,3)
                        auto B_T = Jtinv * DN.transpose(); // matrix of shp fct derivatives in 'x' space; shape=(3, nnods_by_element)

                        Eigen::MatrixXd B_M;
                        switch (ndim)
                        {
                        case 2:
                            B_M.resize(3, elems->type_nnods * ndim); // 3 en 2D; 6 en 3D
                            for (int j = 0; j < elems->type_nnods; ++j)
                            {
                                double ddx = B_T(0, j);
                                double ddy = B_T(1, j);
                                Eigen::MatrixXd Bs(3, 2);
                                Bs << ddx, 0.0,
                                    0.0, ddy,
                                    ddy, ddx;
                                B_M.block(0, ndim * j, 3, 2) = Bs;
                            }
                            break;
                        case 3:
                            B_M.resize(6, elems->type_nnods * ndim); // 3 en 2D; 6 en 3D
                            for (int j = 0; j < elems->type_nnods; ++j)
                            {
                                // double v = basisF(igp, j);
                                double ddx = B_T(0, j);
                                double ddy = B_T(1, j);
                                double ddz = B_T(2, j);
                                Eigen::MatrixXd Bs(6, 3);
                                Bs << ddx, 0.0, 0.0,
                                    0.0, ddy, 0.0,
                                    0.0, 0.0, ddz,
                                    ddy, ddx, 0.0,
                                    ddz, 0.0, ddx,
                                    0.0, ddz, ddy;
                                // std::cout << "v = " << v << '\n';
                                B_M.block(0, ndim * j, 6, 3) = Bs;
                            }
                            break;
                        default:
                            throw std::runtime_error("dimension " + std::to_string(ndim) + " not implemented");
                        }

                        auto eps = B_M * displ;               // nodal epsilon (Voigt)
                        auto sig = medium->hooke(ndim) * eps; // nodal sigma (Voigt)
                        // Eigen::IOFormat fmt(6, 0, ", ", "\n", "\t[", "]");
                        // std::cout << "sig = \n"
                        //           << sig.format(fmt) << '\n';

                        // convert to 3D Voigt vectors
                        Eigen::VectorXd eps3d(6), sig3d(6); // TODO: use fixed-size vectors
                        if (ndim == 3)
                        {
                            eps3d = eps;
                            sig3d = sig;
                        }
                        else
                        {
                            double eps_zz = -medium->m_nu / medium->m_E * (sig[0] + sig[1]);
                            eps3d << eps(0), eps(1), eps_zz,
                                eps(2), 0.0, 0.0;
                            sig3d << sig(0), sig(1), 0.0,
                                sig(2), 0.0, 0.0;
                        }

                        // [ [0]     [ [XX]     [ [0] [3] [4]    [ [XX] [XY] [XZ]
                        //   [1]       [YY]       [3] [1] [5]      [XY] [YY] [YZ]
                        //   [2]       [ZZ]       [4] [5] [2] ]    [XZ] [YZ] [ZZ] ]
                        //   [3]       [XY]
                        //   [4]       [XZ]
                        //   [5] ]     [YZ] ]

                        std::vector<int> voigt3d = {0, 3, 4, 3, 1, 5, 4, 5, 2};
                        for (int i = 0; i < 9; ++i)
                        {
                            eps_nodes.at(inod * 9 + i) = eps3d(voigt3d[i]); // TODO: remove .at => []
                            sig_nodes.at(inod * 9 + i) = sig3d(voigt3d[i]);
                        }

                        //#pragma omp critical
                        {
                            // store data also by node for smoothed fields
                            eps_by_node[ntag].push_back(eps3d);
                            sig_by_node[ntag].push_back(sig3d);
                        }

                    } // for (int inod = 0; ... (nodes of element ie)

                    //#pragma omp critical
                    {
                        etags_all.push_back(etag);
                        epsilon_all.push_back(eps_nodes);
                        sigma_all.push_back(sig_nodes);
                    }

                } // for (size_t ie = 0; ... (elements)
            } // for (auto &elems : entity->esets)

    timers["compute_sig_epl"].stop();
    std::cout << "\telasped = " << timers["compute_sig_epl"].read() << "s\n";
}

FEM_API std::ostream &
operator<<(std::ostream &out, Solver const &obj)
{
    out << "Solver\n";
    return out;
}

} // namespace fem
